/*
//! Async-graphql integration with Warp

#![allow(clippy::type_complexity)]
#![forbid(unsafe_code)]

pub mod batch_request;
pub mod errors;
pub mod requests;
pub mod subscriptions;
pub mod filters;

pub use batch_request::{graphql_batch, graphql_batch_opts, BatchResponse};
pub use errors::BadRequest;
pub use requests::{graphql, graphql_opts, Response};
// pub use subscriptions::{graphql_subscription, graphql_subscription_with_data};
*/

use async_graphql::scalar;
use async_graphql::{EmptyMutation, EmptySubscription, MergedObject, Schema, Enum};
// use async_graphql_warp::{BadRequest, Response};
use async_graphql_warp::{Response};
use serde::{Serialize,Deserialize};
//use http::StatusCode;
use std::convert::Infallible;
use warp::{
  filters::header::headers_cloned,
  filters::method::method,
  http::{method::Method, HeaderMap, HeaderValue},
  // http::{header::AUTHORIZATION, method::Method, HeaderMap, HeaderValue},
  // filters::path::{full, FullPath},
  http::Response as HttpResponse,
  Filter, filters::BoxedFilter
};

use gql_playground::{playground_source,GraphQLPlaygroundConfig};
	//use async_graphql::http::{playground_source, GraphQLPlaygroundConfig};
  //use crate::topographic_anatomy::graphql::{QueryRoot as TaQueryRoot};
// use crate::topographic_anatomy::{TopographicAnatomyQuery,TopographicAnatomyMutation}; // TopographicAnatomyChanged};
// use crate::tracking_point::{TrackingPointQuery,TrackingPointMutation}; // TrackingPointChanged};
  // use zterton::auth::defs::{AuthStore};
  // use zterton::auth::filters::{with_auth};
use crate::defs::{DataDBs};
use reqenv::ReqEnv;

#[derive(Enum, Eq, PartialEq, Copy, Clone)]
pub enum MutationType {
    Created,
    Updated,
    Deleted,
}
impl Default for MutationType {
  fn default() -> Self { MutationType::Updated } 
}

pub type JsonMap = serde_json::Map<String, serde_json::Value>;

#[derive(MergedObject, Default)]
struct Query(
//   TopographicAnatomyQuery,
//   TrackingPointQuery,
);

#[derive(MergedObject, Default)]
struct Mutation(
//   // TopographicAnatomyChanged,
//   TopographicAnatomyMutation,
//   // TrackingPointChanged,
//   TrackingPointMutation,
);

#[derive(Clone, Default, Debug, Serialize, Deserialize)]
pub struct Map {
  pub data: JsonMap,
}
scalar!(Map);

pub async fn graphiql(
  path: &'static str,
  giql_path: &'static str,
  data_dbs: DataDBs,
) -> BoxedFilter<(impl warp::Reply,)> {
// ) -> impl Filter<Extract = impl warp::Reply, Error = warp::Rejection> + Clone {
   if giql_path.is_empty() {
    warp::path(path)
    .and(warp::get())
    .map(move || {
        HttpResponse::builder()
          .header("content-type", "text/html")
          .body("".into())
      })
      .boxed()
  } else {
   warp::path(path)
   .and(warp::get())
   .and(headers_cloned())
   .and(method())
   .and(with_db(data_dbs))
   // TODO  check auth here
   .map(move |_headers: HeaderMap<HeaderValue>, _method: Method, _db: DataDBs| {
      let gql_playground_config: GraphQLPlaygroundConfig = GraphQLPlaygroundConfig::new(giql_path);
      HttpResponse::builder()
        .header("content-type", "text/html")
        .body(playground_source(gql_playground_config))
//        .body(playground_source(GraphQLPlaygroundConfig::new("/gql")))
    })
    .boxed()
  }

//    graphql_playground
//        .or(graphql_post)
    //     .recover(|err: Rejection| async move {
    //         if let Some(BadRequest(err)) = err.find() {
    //             return Ok::<_, Infallible>(warp::reply::with_status(
    //                 err.to_string(),
    //                 StatusCode::BAD_REQUEST,
    //             ));
    //         }
    //         Ok(warp::reply::with_status(
    //             "INTERNAL_SERVER_ERROR".to_string(),
    //             StatusCode::INTERNAL_SERVER_ERROR,
    //         ))
    // })
}


pub async fn graphql(
//  config: &Config,
  path: &'static str,
  data_dbs: DataDBs,
  cors: warp::cors::Builder,
) -> BoxedFilter<(impl warp::Reply,)> {
// ) -> impl Filter<Extract = impl warp::Reply, Error = warp::Rejection> + Clone {

    let schema = Schema::build( Query::default(), Mutation::default(), EmptySubscription)
        .data(data_dbs.clone())
        .finish();

    // warp::path!("gql").and(async_graphql_warp::graphql(schema)).and_then(
      warp::path(path)
      .and(headers_cloned())
      .and(method())
      .and(with_db(data_dbs))
      .and(async_graphql_warp::graphql(schema)).and_then(
        | headers: HeaderMap<HeaderValue>, method: Method, db: DataDBs,
        (schema, request): (
            Schema<Query, Mutation, EmptySubscription>,
            async_graphql::Request,
        )| async move { 
          let reqenv = ReqEnv::new(db.app, db.auth, headers, method, "/gql", "gql", "gql");
          match reqenv.user_authentication().await {
            Ok(_auth) => {
              // dbg!("auth: {}",&auth);
              Ok::<_, Infallible>(Response::from(schema.execute(request).await))
            },
            Err(e) => {
              let result = format!("Error: no credentials found");
              println!("gql:{} {}",&result,e);
              let empty_schema = Schema::build( Query::default(), EmptyMutation, EmptySubscription)
              .finish();
              Ok::<_, Infallible>(Response::from(empty_schema.execute(request).await))
            }
          }
        }
      )
      .with(cors)
      .boxed()
}

fn with_db(db: DataDBs) -> impl Filter<Extract = (DataDBs,), Error = std::convert::Infallible> + Clone {
    warp::any().map(move || db.clone())
}